#include "Projectile.h"
#include "MathExt.h"
#include <cmath>
#include <cassert>


//---------------------------------------------------------------------------------------------
//

CProjectile::CProjectile(double x0, double y0, double yf, double vi, double deg)
: x0_(x0), y0_(y0), yf_(yf), vx0_(vi * ::cos(CMathExt::toRad(deg))), vy0_(vi * ::sin(CMathExt::toRad(deg))), g_(9.81)
{}

double CProjectile::getElapsedTime() const
{
	//  [ 15-17b] ߷¸  ߻ü y      ҿð
	//  [ 15-11], [ 15-15], [ 15-20], [ 15-26] [ 15-32]    ̴

	//  ȣ  0 Ŀ Ѵ.
	if (vy0_ * vy0_ - 2.0 * g_ * (yf_ - y0_) > 0.0)
		return (vy0_ + ::sqrt(vy0_ * vy0_ - 2.0 * g_ * (yf_ - y0_))) / g_;
	//  error Ȳ
	else return 0.0;
}

double CProjectile::getVx(double/* t*/) const
{
	//  [ 15-6a] ߷¸  ߻ü x ӵ 
	return vx0_;
}

double CProjectile::getVy(double t) const
{
	//  ð   ܻ ó
	if (t <= 0.0) return vy0_;
	//  Ѽҿð Ŭ  ӵ ˷ش
	else if (t >= getElapsedTime())
		t = getElapsedTime();

	//  [ 15-6b] ߷¸  ߻ü y ӵ 
	return -g_ * t + vy0_;
}

double CProjectile::getX(double t) const
{
	//  ð   ܻ ó
	if (t <= 0.0) return x0_;
	//  Ѽҿð Ŭ  ġ ˷ش, [ 15-9]
	else if (t >= getElapsedTime()) return vx0_ * getElapsedTime() + x0_;

	//  [ 15-8a] ߷¸  ߻ü x  
	return vx0_ * t + x0_;
}

double CProjectile::getY(double t) const
{
	//  ð   ܻ ó
	if (t <= 0.0) return y0_;
	//  Ѽҿð Ŭ  ġ ˷ش
	else if (t >= getElapsedTime()) return yf_;

	//  [ 15-8b] ߷¸  ߻ü y  
	return -0.5 * g_ * t * t + vy0_ * t + y0_;
}


//---------------------------------------------------------------------------------------------
//

CProjectileWithDragForce::CProjectileWithDragForce(double x0, double y0, double yf, double vi, double deg, double m, double c)
: CProjectile(x0, y0, yf, vi, deg), m_(m), c_(c)
{}

double CProjectileWithDragForce::getElapsedTime() const
{
	//  [ 15-42] ׷  ߻ü       ҿð
	double seedTime = CProjectile::getElapsedTime();
	double elapsedTime = solveElapsedTime(seedTime);
	assert(elapsedTime > 0.0);
	return elapsedTime;
}

double CProjectileWithDragForce::getVx(double t) const
{
	//  ð   ܻ ó
	if (t <= 0.0) return vx0_;
	//  Ѽҿð Ŭ  ӵ ˷ش
	else if (t >= getElapsedTime())
		t = getElapsedTime();

	//  [ 15-39] ׷  ߻ü  ӵ 
	return vx0_ * ::exp(-t * c_ / m_);
}

double CProjectileWithDragForce::getVy(double t) const
{
	//  ð   ܻ ó
	if (t <= 0.0) return vy0_;
	//  Ѽҿð Ŭ  ӵ ˷ش
	else if (t >= getElapsedTime())
		t = getElapsedTime();

	//  [ 15-41] ׷  ߻ü  ӵ 
	return (vy0_ + m_ * g_ / c_) * ::exp(-t * c_ / m_) - m_ * g_ / c_;
}

double CProjectileWithDragForce::getX(double t) const
{
	//  ð   ܻ ó
	if (t <= 0.0) return x0_;
	//  Ѽҿð Ŭ  ӵ ˷ش
	else if (t >= getElapsedTime())
		t = getElapsedTime();

	//  [ 15-40] ׷  ߻ü   
	return vx0_ * (m_ / c_) * (1.0 - ::exp(-t * c_ / m_));
}

double CProjectileWithDragForce::getY(double t) const
{
	//  ð   ܻ ó
	if (t <= 0.0) return y0_;
	//  Ѽҿð Ŭ  ӵ ˷ش
	else if (t >= getElapsedTime())
		t = getElapsedTime();

	//  [ 15-42] ׷  ߻ü   
	return (vy0_ + m_ * g_ / c_) * (m_ / c_) * (1.0 - ::exp(-t * c_ / m_)) - t * m_ * g_ / c_ + y0_;
}

//  ġؼ  : secant method
double CProjectileWithDragForce::solveElapsedTime(double init) const
{
	double tolerance = CMathExt::EPS;
	
	double final = init + 2.0;
	double front = calcElapsedTime(init);
	double rear = calcElapsedTime(final);
	double delta;
	
	int i = 0;
	do {
		if (-tolerance <= rear-front && rear-front <= tolerance) {
			assert(false);
			return 0.0;
		}

		delta = -rear * (final - init) / (rear - front);
		init = final;
		final += delta;
		front = rear;
		rear = calcElapsedTime(final);
		++i;

		if (i >= 1000) {
			assert(false);
			return final;
		}
	} while (::fabs(delta) > tolerance || ::fabs(rear) > tolerance);
	
	return final;
}

double CProjectileWithDragForce::calcElapsedTime(double t) const
{
	return (vy0_ + m_ * g_ / c_) * (m_ / c_) * (::exp(-t * c_ / m_) - 1.0) + t * m_ * g_ / c_ + yf_;
}
